/*
  Servo360driver.c  ver 11/11/2017
                    Corrected passing arrays in functions
       Based on a Spin/PASM driver by Tom Crawford
     
     Group sections in this order:
      Demo global declarations
      Driver global declarations  
      Driver functions
      Demo functions
      main()
      
*/
#include "simpletools.h"                      // Include simple tools
#include "stdint.h"

//  ======================================================
//       DEMO defines, variables, function prototypes
//  ======================================================
// **** defines for demo
// ***  Servo pins control and feedback, if both 0 -- no servo
#define CP0 12
#define FP0 8
#define CP1 13
#define FP1 9
#define CP2 14
#define FP2 10
#define CP3 0
#define FP3 0
// ***  servo names
#define Servo0 0
#define Servo1 1
#define Servo2 2
#define Servo3 3

// **** Declare demo globals
//int8_t  Cog, DisplayServo, currentservo, testgain,
volatile int8_t  deltaMax, deltaMin;  
volatile int16_t ordered, actual, rcount, theservo, Cmnded[4];    // short sub ordered for commanded
int random;

// **** Demo function prototypes
void MySet(short, short);       //  servonum, degrees
short WaitSet(short);            //  servonum,  returns position in deg
void showangles();

//  ======================================================
//       DRIVER variables, function prototypes
//  ======================================================

//  **** Declare DRIVER globals  
int MyParm[10];               // passes addresses of parameters
volatile int8_t mode[4], Complete[4];
volatile int16_t ServoPins[4];         //  must be [4], ControlPin<<8 | FeedbackPin
volatile int16_t PulseWidth[4], theta[4], commanded[4], index ;
volatile int Context[4], Parameters[4]; 
volatile int *cog360servo;

// **** DRIVER function prototypes
void start360(int *MyParm);  
void SetPos(short, short);          // theservo, commanded[theservo] in degrees
void Width(short, short);           // theservo, microsecs
short GetPos(short);                // theservo, returns theta in degrees
int8_t CheckDone(short);            // theservo, returns Complete 0 at position, !=0 if not
void SetMaxError(short, int8_t);    // theservo, error +1 in degrees
void SetGain(short, int8_t);        // theservo, gainP (5 to 15)
void SetDeltaMin(short, int8_t);    // theservo, Dmin  (approx 30)
void SetDeltaMax(short, int8_t);    // theservo, Dmax (approx 150)

//  ======================================================
//            START DRIVER FUNCTIONS
//  ======================================================

//  ***************************
//      Initialize
//  ***************************
void start360(int *MyParm)     // pin numbers from define in header
{
 ServoPins[0] = CP0 << 8 | FP0;
 ServoPins[1] = CP1 << 8 | FP1;
 ServoPins[2] = CP2 << 8 | FP2;
 ServoPins[3] = CP3 << 8 | FP3;
 for( int i = 0; i <= 3; i++)
   {
     if (ServoPins[i] != 0)
      {
        mode[i] = 0;
        PulseWidth[i] = 1500;                         
        Parameters[i] = (2<<24)|(18<<16)|(31<<8)|150;    // error-3, gain-12, deltamin-31, deltamax-150
      } // end if
    }   // end for i
 MyParm[0] = PulseWidth;
 MyParm[1] = mode; 
 MyParm[2] = theta;
 MyParm[3] = commanded; 
 MyParm[4] = Complete;        
 MyParm[5] = CLKFREQ / 1000000;
 MyParm[6] = ServoPins; 
 MyParm[7] = Context;
 MyParm[8] = Parameters;

 extern int binary_FourPly_dat_start[]; 
 cog360servo = cognew((void*)binary_FourPly_dat_start, MyParm); 
 pause(500);          
 }
 
//  ******************************************************
//        Order servo to position - degrees 
//  ******************************************************
void SetPos(short servonum, short degrees)
{
  commanded[servonum] = degrees;
  Complete[servonum] = 1;                  //we are not complete yet
  mode[servonum] = 1;                      //set position mode
 }
 
//  ******************************************************
//      Make servo receive pulses of spec'd width
//  ****************************************************** 
void Width(short servonum, short Usec)
{
  mode[servonum]= 0;                //  CPW mode
  PulseWidth[servonum] = Usec;      // save pulse width in usec
 }
 
//  ******************************************************
//      Get the current position of the servo
//  ****************************************************** 
short GetPos(short servonum)
{
  int16_t xxx = theta[servonum];
  return xxx;
  //return theta[servonum];    // can't return an array element
}  

//  ******************************************************
//     Check if servo is at commanded position
//     Returns Complete[servonum]
//       =0 if in commanded position +/- error for 200ms
//      !=0 if not in position 
//  ******************************************************  
int8_t CheckDone(short servonum)
{
  int8_t xxx = Complete[servonum];
  return xxx;
  //return Complete[servonum];    // can't return an array element
}  

//  ******************************************************
//    Define the maximum allowable error for spec'd servo
//      accepterror = 1 degree greater than acceptable error
//  ****************************************************** 
void SetMaxError(short servonum, int8_t accepterror)
{
  Parameters[servonum] = Parameters[servonum] & 0x00FFFFFF;
  Parameters[servonum] = Parameters[servonum] | (accepterror << 24);  
}

//  ******************************************************
//   Define the gain for the specified servo. 
//    typical 5 to 15, higher results in more overshoot   
//  ******************************************************
void SetGain(short servonum, int8_t gainP)
{
  Parameters[servonum] = Parameters[servonum] & 0xFF00FFFF;  
  Parameters[servonum] = Parameters[servonum] | (gainP << 16);
}

//  ******************************************************
//   Specifies the minimum deviation from 1500 usec 
//    Use to keep servo from being stuck in dead zone
//    Typical values approx 30  
//  ******************************************************
void SetDeltaMin(short servonum, int8_t Dmin) 
{
  Parameters[servonum] = Parameters[servonum] & 0xFFFF00FF;  
  Parameters[servonum] = Parameters[servonum] | (Dmin << 8);  
}

//  ******************************************************
//    Specifies the maximum deviation from 1500 usec.
//      Typical value approx 150
//  ******************************************************
void SetDeltaMax(short servonum, int8_t Dmax)
{
  Parameters[servonum] = Parameters[servonum] & 0xFFFFFF00;  
  Parameters[servonum] = Parameters[servonum] | Dmax;   
}
//  ======================================================
//        END DRIVER FUNCTIONS 
//  ======================================================

  
//  ======================================================
//        Start DEMO FUNCTIONS 
//  ======================================================
//  ******************************************************
//         Set ordered position
//  ******************************************************
void MySet(short servonum, short degrees)
{
  SetPos(servonum, degrees);
}   

//  ******************************************************
//    Wait until servo has completed ordered movement     
//  ******************************************************
short WaitSet(short servonum)
{
  while(CheckDone(servonum));
  pause(10);
  return GetPos(servonum);
}

//  ******************************************************
//    Print servo position in degrees     
//  ******************************************************
void showangles()
{
  short servonum;
  for(servonum = 0; servonum <= 2; servonum++)
    {
       print("  Servo%d %d  ",servonum, GetPos(servonum));
    }        
}
//  ******************************************************
//        END DEMO FUNCTIONS 
//  ******************************************************
//  ******************************************************

int main()                   // The DEMO
{
 start360(MyParm);
 print("cog360 = %d \n", cog360servo);
 pause(500);

// Parameters values set for all servos in start360() 
// Parameters[i] = (2<<24)|(18<<16)|(31<<8)|150;    // in spin: error-3, gain-12, deltamin-31, deltamax-150 

// ***** Set parameters for individual servos 
//SetMaxError(Servo0, 4);
//SetMaxError(Servo1, 4); 
//SetMaxError(Servo2, 4);
 pause(10);
 
 SetGain(Servo0, 16); // 22
 SetGain(Servo1, 12); // 24
 //SetGain(Servo2, 20);
  pause(10); 
 
SetDeltaMin(Servo0, 29);
 SetDeltaMin(Servo1, 29);
 SetDeltaMin(Servo2, 31);
 pause(10);
//  ***** End parameter set*****

//  ***** command servos to move *****
 Width(Servo0, 1475);   // start servos rotating 1400 - 1600
 Width(Servo1, 1530);
 Width(Servo2, 1450);
 Width(Servo3, 1550);
 
 for(int i = 0; i < 50; i++)
  {
    showangles();
    print("\n");
    pause(10);
  }       
 for(theservo = 0; theservo <= 2; theservo++) Width(theservo,1500);   // stop rotation
 pause(4000);
 
 for(theservo = 0; theservo <= 2; theservo++)   // full scale changes
  {
    print("servo %d  ",theservo);
    MySet(theservo, 180);
   // pause(1000);
    //actual = GetPos(theservo);
    actual = WaitSet(theservo);
    print("%d  ",actual);
    MySet(theservo, 0);
   // pause(1000);
    actual = WaitSet(theservo);
    //actual = GetPos(theservo);
    print("%d  ",actual);    
    MySet(theservo, 358);
   // pause(1500);
    actual = WaitSet(theservo);    
    //actual = GetPos(theservo);
    print("%d  ",actual);    
    MySet(theservo, 0);
   // pause(1200);
    actual = WaitSet(theservo);
    //actual = GetPos(theservo);
    print("%d  \n",actual);
  }                             // end full scale changes
  
  srand(CNT);
  random = rand();
  rcount = 0;
  while(rcount < 100)
    {
      for(theservo = 0; theservo <= 2; theservo++)   // 
      {
       if(CheckDone(theservo) == 0)
        {
         print("\n%d  Commanded: %3d  Actual: %3d", theservo, Cmnded[theservo], GetPos(theservo));
         print("  Error: %2d", Cmnded[theservo] - GetPos(theservo));
         
         random = rand();
         ordered = abs(random) % 360;
         MySet(theservo, ordered);
         pause(200);
         Cmnded[theservo] = ordered;
         print("   %2d", rcount++);
        }   // end if      
      }     // end for 
    }       // end while
}   // end main

